"""
__init__ module for custom DataTypes created by fjformer.core
"""

from fjformer.dtypes.arraynf4 import ArrayNF4 as ArrayNF4
from fjformer.dtypes.array8bit import Array8Bit as Array8Bit
from fjformer.dtypes.array8lt import Array8Lt as Array8Lt
